##############################################################
# This file runs sklearn RF and NN
##############################################################


import numpy as np
import os
from sklearn.metrics import roc_curve, roc_auc_score, brier_score_loss
from sklearn.neural_network import MLPClassifier

import pickle


#==============================
# Functions
#==============================


#=============================
# Get the current date chunk
#=============================

def getDateChunk(in_dates,k, curr_chunk,ndays1,ndays2,nChunks1,nChunks2, daySplit):
# in_dates is the incoming array of dates (an array of strings)
# k is the total number of chunks
# curr_chunk is the current chunk for testing
# ndays1 is the number of days in chunk/group 1
# ndays2 is the number of days in chunk/group 2
# nChunks1 is the number of chunks in group1
# nChunks2 is the number of chunks in group2
# daySplit is the list of day indices telling how in_dates should be split

#Split the dates array into chunks. The first 4 chunks have 29 days;
#The last 3 chunks have 30 days

#Obtain testing, validation, and training dates for current fold.
#How this works: testing is the current "chunk", 
#validation is the next "chunk", and training comprises all other chunks

    date_chunks = np.array_split(in_dates, daySplit)

    test_date = date_chunks[curr_chunk]

    #if the last chunk is the testing chunk
    if (curr_chunk == k-1):
        #val_date = date_chunks[0]
        train_date = np.concatenate(date_chunks[0:curr_chunk],axis=0)

    #if first chunk is the testing chunk
    elif (curr_chunk == 0):
        train_date = np.concatenate(date_chunks[1:],axis=0)

    else:
        #val_date = date_chunks[curr_chunk + 1]
        train_date1 = date_chunks[:curr_chunk]
        train_date2 = date_chunks[curr_chunk + 1:]
        train_date = train_date1 + train_date2
        train_date = np.concatenate((train_date), axis=0)


    return test_date, train_date


#====================================
# Get necessary info to make splits
#====================================

#Will obtain the list which shows how the data should be split
#(given uneven chunks)
def get_split_info(sampPerChunk1, sampPerChunk2,n1,n2):
#sampPerChunk1 is the samples_per_chunk1 variable (user defined)
#sampPerChunk2 is the samples_per_chunk2 variable
#n1 is the number of chunks in group 1
#n2 is the number of chunks in group 2
    chunk_list = []

    #first chunk
    for c in range(1,n1+1):
        chunk_list.append(c*sampPerChunk1)

    if (n2 > 0): #make sure we indeed have two chunks; if not, the follwoing code doesn't make sense
    #second chunk
        for c in range(1,n2): # only n2 instead of n2+1 b/c of how np.array_split works
            chunk_list.append(n1*sampPerChunk1+c*sampPerChunk2)

    return chunk_list



def read_binary(infile, dx_pts, dy_pts, header):
    #@infile is the filename of unformatted binary file
    #Returns a numpy array of data
    #@dx_pts is an integer describing the number of longitude points
    #@dy_pts is an integer describing the number of latitude points
    #@header is a binary variable -- True if contains a 1-elmnt header/footer at 
    #beginning and end of file; else False


    f = open ( infile , 'rb' )
    arr = np.fromfile ( f , dtype = np.float32 , count = -1 )
    f.close()

    if (header == True):
        arr = arr[1:-1]


    #print arr.shape
    #Rebroadcast the data into 2-d array with proper format
    arr.shape = (dy_pts, dx_pts)
    return arr

def read_obs_binary(infile):
    #@infile is the filename of unformatted binary file
    #Returns a numpy array of data

    f = open ( infile , 'rb' )
    arr = np.fromfile ( f , dtype = np.float32 , count = -1 )
    f.close()

    return arr

#Returns an array containing (y,x) grid points in the mask 
#mask_var is a np array of binary mask points (1 in the mask; else 0)
#nx is total number of x grid points, ny is total num of y grid points
def get_domain_pts(mask_var, nx, ny):

    mask_pts = []
    for i in range(nx):
        for j in range(ny):
            if (mask_var[j,i] >= 0.9):
                mask_pts.append((j,i))

    mask_pts = np.array(mask_pts)

    return mask_pts

#Reformats the data to a format acceptable to sklearn
#i.e., going from (samples, y, x, channels) to (samples, variables)
#nsize is neighborhood size (e.g., 5 means consider 5x5 neighborhood around the center point)
def reformat_data(in_data, nsize, n_samples, fdate, center_pt, radius):

    #variables will be as follows:
    #channel1 at -5 to +5 from center point, channel 2 at -5 to +5, etc. 

    #n_vars = nsize*nsize*n_channels + 12#total # of vars considered by sklearn
    #new_data = np.zeros((n_samples, n_vars))

    #Extract just the relevant "patches"
    #new_data = in_data[:,center_pt-radius:center_pt+radius+1, center_pt-radius:center_pt+radius+1, 0:-2]
    
    new_data = in_data[:,center_pt-radius:center_pt+radius+1, center_pt-radius:center_pt+radius+1, :]

    #But how to handle the lat/lon -- only really want lat/lon at the center point
    #lats = in_data[:,center_pt, center_pt, -2]
    #lons = in_data[:, center_pt, center_pt, -1] 

    #lats.shape = (len(lats), 1) 
    #lons.shape = (len(lons), 1) 

    #reshape to a 2d array

    #order: spans through the variables first and then the points
    new_data = new_data.reshape(n_samples, -1)


    #Add the lats and lons
    #new_data = np.concatenate( (new_data, lats), axis=1)
    #new_data = np.concatenate((new_data, lons), axis=1)

    
    return new_data


#Returns a vector of the daily samples [shape (samples, height, width, channels)]
#and the daily observations [i.e., of shape (samples)]. 
#@day_fcst is the day's forecast vector [shape (dy,dx,n_channels)]
#@day_obs is the day's observation vector [shape (dy,dx)]
#@mask_pts is the set of domain points [shape (dy,dx)]
#@size_patch is the size of the width/height of the image (i.e., a scalar)
#@n_chan is the number of channels
#@samp_rate is the number of points to randomly sample
#@llats is the latitudes (j,i)
#@llons is the longitudes (j,i)

#TODO: implement samp_rate
def get_patch_data(day_fcst, day_obs, mask_pts, size_patch, n_chan, samp_rate, llats, llons):

    if (samp_rate < 1.):
        n_rand_samples = int(len(mask_pts)*samp_rate)
    else:
        n_rand_samples = len(mask_pts)

    samples = np.zeros((n_rand_samples, size_patch, size_patch, n_chan))
    obs_list = np.zeros(n_rand_samples)
    point_radius = (size_patch)//2 #radius from center point

    #New approach: to limit the amount of samples (memory issues): maybe sample 10% of the points 
    #Step 1: Randomly sample the points in mask_pts 

    if (samp_rate < 1.):
        rand_inds = np.random.choice(np.arange(len(mask_pts)), size=n_rand_samples, replace=False)
        sample_mask_pts = mask_pts[rand_inds]

    else:
        sample_mask_pts = mask_pts

    #span the number of points in the analysis domain
    #for each point: get the obs and get the 101 x 101 x n-channels image
    for i in range(len(sample_mask_pts)):
        #set the x and y points
        #print (sample_mask_pts[i])

        ypt = sample_mask_pts[i,0]
        xpt = sample_mask_pts[i,1]
        #print (ypt, xpt)

        #get the obs
        obs_list[i] = day_obs[ypt, xpt]

        #get the lat and lon
        lat_field = llats[ypt-point_radius:ypt+point_radius+1, xpt-point_radius:xpt+point_radius+1]
        lon_field = llons[ypt-point_radius:ypt+point_radius+1, xpt-point_radius:xpt+point_radius+1]


        #Original (i.e., DON'T express surrounding grid points in relative framework:
        #get the forecast patch
        for c in range(n_chan):
            #print (ypt, xpt)
            if (c < n_chan - 2): #if not last two channels:
                samples[i, :,:, c] = day_fcst[ypt-point_radius:ypt+point_radius+1,xpt-point_radius:xpt+point_radius+1,c]
            elif (c == n_chan -2): #the lat/lon fields will be the last two channels
                samples[i, :,:, c] = lat_field
            elif (c == n_chan -1 ):
                samples[i, :,:, c] = lon_field

        #New (i.e., DO express surrounding grid points in relative framework: -- NOTE: Didn't work as well
        #for c in range(n_chan):
        #    if (c < n_chan - 2): #if not last two channels:
        #        samples[i, :,:, c] = day_fcst[ypt-point_radius:ypt+point_radius+1,xpt-point_radius:xpt+point_radius+1,c] - day_fcst[ypt, xpt, c]
        #        #set the center point (b/c now will be zero)
        #        samples[i,point_radius, point_radius,c] = day_fcst[point_radius, point_radius, c]
        #    elif (c == n_chan -2): #the lat/lon fields will be the last two channels
        #        samples[i, :,:, c] = lat_field
        #    elif (c == n_chan -1 ):
        #        samples[i, :,:, c] = lon_field



    return samples, obs_list


#Does the preprocessing: outputs individual-day .dat files to 
#sepcified directory to be used later for kfold cross-validation
#centerPt is the center point of the patch 
#rad is the RF radius 
def do_preprocessing(fdates, ddx, ddy, n_chan, stat_list, var_list, fdir, odir, raw_dir, domain_pts, size_patch, samp_rate, lats, lons, n_samples, nsize, outdir_fcst, outdir_obs, centerPt, rad):

    for d in range(len(fdates)):

        #daily_fcst = np.zeros((n_mask_pts, n_vars))
        daily_fcst = np.zeros((ddy, ddx, n_chan)) 
        fdate = fdates[d]
        print (fdate) 

        ffiles = ["%s/sseo_ens_%s_%s_80km_%s00.dat" %(fdir, stat, var, fdate) for stat in stat_list for var in var_list]

        #Add the raw probability file
        #NOTE: Will have to consider whether we want to actually include this as a predictor in the ML or just use for comparison
        #For now we'll include it b/c of convenience, but it's something to think about 
        #raw_baseline_file = "%s/%s_UH30_ens_PROBS.dat" %(raw_dir, fdate)
        raw_baseline_file = "%s/%s_UH%s_ens_PROBS.dat" %(raw_dir, fdate, uhthresh) #maximizes AUC


        #Try not adding in raw baseline...for now
        ffiles.append(raw_baseline_file)

        ofile = "%s/%s00_REPS211_hail.dat" %(odir, fdate)

        #Read in the forecast and obs data
        daily_obs = read_binary(ofile, ddx, ddy, False)

        raw_baseline = read_binary(raw_baseline_file, ddx, ddy, False)
        for c in range(len(ffiles)):
            daily_fcst[:,:,c] = read_binary(ffiles[c], ddx, ddy, False)

        print ("after read daily forecast")

        #Implement the patches algorithm to obtain each day's samples
        daily_samples, daily_patch_obs = get_patch_data(daily_fcst, daily_obs, domain_pts, size_patch, n_chan, samp_rate, lats, lons)

        print ("after get_patch")
        print (daily_samples.shape)
        print (daily_patch_obs.shape)

        #TODO: Now is when you'd want to reformat the data; then you can save it to .dat file
        #TODO: Not done updating below--
        new_daily_samples = reformat_data(daily_samples, nsize, n_samples, fdate, centerPt, rad)

        #daily_patch_obs is already in the proper format (shape = samples) 

        print (new_daily_samples.shape) 
        print (daily_patch_obs.shape) 
        new_daily_samples = np.float32(new_daily_samples)
        daily_patch_obs = np.float32(daily_patch_obs)

        #Write to file 
        new_daily_samples.tofile("%s/%s_fpatch.dat" %(outdir_fcst, fdate))
        daily_patch_obs.tofile("%s/%s_obs.dat" %(outdir_obs, fdate) )

    return

#Saves individual-day probabilities to .txt file 
#(saves over entire domain; not just mask points) 
#mask_probs is the probs within the analysis domain (1D) from a single day 
#ddx andd ddy are the number of x and y points, respectively in the full area (not just the mask) 
#mask_pts is an array of the points within the domain area (y,x)
#outdir is the output directory for the txt file
#fdate is the date in question
def save_probs(mask_probs, ddx, ddy, mask_pts, outdir, fdate):

    output_probs = np.ones((dy,dx))*-999.
    
    for l in range(len(mask_probs)):
        y_pt = mask_pts[l,0]
        x_pt = mask_pts[l,1]
        output_probs[y_pt, x_pt] = mask_probs[l]


    np.savetxt("%s/%s_rf_probs.txt" %(outdir, fdate), output_probs)

    return 



#==================================
# User defined parameters
#==================================

dx = 93 #num x points
dy = 65 #num y points

variables = ["REFDHM", "RELH", "TMPK", "DWPK", "P01M", "UPDRHM", "CAPE", "SHEAR", "UPHLHM", "CAPESHEAR" ]

test_date_file = "test_dates.txt"
train_date_file = "train_dates.txt" 

n_mask_pts = 1434
n_size_rf = 1
n_vars = (len(variables)+2)*n_size_rf*n_size_rf # +2 because adding latitude and longitude

n_channels = len(variables)+2

sampling_rate = 1.0

#RF hyperparameters
#n_trees = 200
#best_max_depth = 15
#best_min_samples_leaf = 20
#n_proc = 36

radius = n_size_rf//2 #radius from center point
patch_size = n_size_rf
center_pt = patch_size//2

n_days = 221

samples_per_day = int(n_mask_pts*sampling_rate)
total_samples = int(n_days*samples_per_day)

clf_filename = "NN_hail_tutorial.pkl"

maskfile = "new_conus_mask_80km.dat"

lat_file = "lat_80km.dat"
lon_file = "lon_80km.dat"

indir_forecast = "../all_x"
indir_observations = "../all_y"

outdir_forecast = "../rf/fcst"
outdir_observations = "../rf/obs"
outdir_txt = "./outfiles" 


#====================================================

#Step 1: Read in the dates, lats, lons; get mask points, etc. 

#Read in the 20km lat/lon points
lats = read_binary(lat_file, dx, dy, True)
lons = read_binary(lon_file, dx, dy, True)

domain_mask = read_binary(maskfile, dx, dy, True)

#Obtain x and y points/coordinates that fall within the domain
domain_pts = get_domain_pts(domain_mask, dx, dy)

#Step 2:
#After preprocessing: get the test/train date split 

test_dates = np.genfromtxt(test_date_file, dtype='str') 
train_dates = np.genfromtxt(train_date_file, dtype='str') 

#Approach: Concatenate the training dates (fcst/obs) into training.dat and testing dates into testing.dat 
#Then read in the .dat files 
#Then train on the training set, test on the testing set -- for testing just output probability files

print ("cancatenating files") 
for d in train_dates:
    ffile = "%s/%s_fpatch.dat" %(outdir_forecast, d) 
    ofile = "%s/%s_obs.dat" %(outdir_observations, d) 
    os.system('cat %s >> training_fdata.dat' %ffile)
    os.system('cat %s >> training_odata.dat' %ofile) 

for d in test_dates:
    ffile = "%s/%s_fpatch.dat" %(outdir_forecast, d)
    ofile = "%s/%s_obs.dat" %(outdir_observations, d)      
    os.system('cat %s >> testing_fdata.dat' %ffile)
    os.system('cat %s >> testing_odata.dat' %ofile)

print ("Reading in test/train data") 
#Read in the data
training_x = read_binary('training_fdata.dat', n_vars,samples_per_day*len(train_dates), False) 
testing_x = read_binary('testing_fdata.dat', n_vars, samples_per_day*len(test_dates), False) 

training_y = read_obs_binary('training_odata.dat') 
testing_y = read_obs_binary('testing_odata.dat') 


#NOTE: Important step: Need to tornamlize the data

train_mean = training_x.mean(axis=0)
train_sd = training_x.std(axis=0) 

#Want to prevent divide by zero error
delta = 0.000001
train_sd += delta

np.save("train_mean.npy", train_mean)
np.save("train_std.npy", train_sd)

training_x = (training_x - train_mean)/train_sd
testing_x = (testing_x - train_mean)/train_sd

print ("Begin training NN") 
#Train the NN
nn = MLPClassifier(hidden_layer_sizes=(64, 12, ), activation='relu', solver='adam', alpha=0.00001, batch_size='auto',\
            learning_rate='adaptive', learning_rate_init=0.0001, max_iter=100, shuffle=True, random_state=None,\
            tol=0.0001, verbose=True, warm_start=True, momentum=0.8, early_stopping=False,\
            validation_fraction=0.2, beta_1=0.9, beta_2=0.999, epsilon=1e-08, n_iter_no_change=30)


nn.fit(training_x, training_y)
print ("nn has been fit") 

#Use pickle to save the NN
nn_pkl = open(clf_filename, 'wb') 
pickle.dump(nn, nn_pkl)
nn_pkl.close()

#Use the nn to make predictions on the testing data
clf_probs = nn.predict_proba(testing_x)[:,1]

#Save individual-day clf_probs to .dat or .txt 
indiv_day_probs = np.array_split(clf_probs, len(test_dates)) 

for d in range(len(indiv_day_probs)):
    print (indiv_day_probs[d].shape)
    date = test_dates[d] 

    save_probs(indiv_day_probs[d], dx, dy, domain_pts, outdir_txt, date)

#At the end of the fold, make sure to remove the files
os.system('rm training*.dat ; rm testing*.dat') 
