#####################################################
# Evaluates RF forecasts 
#####################################################

#============================
# Imports
#============================

from mpl_toolkits.basemap import Basemap
import numpy as np
import modelEvaluationFunctions_new as Eval
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from sklearn.metrics import roc_curve, roc_auc_score, brier_score_loss

import graph_functions as graph

#==============================


#======================
# Functions
#======================

def read_binary(infile, dx_pts, dy_pts, header):
    #@infile is the filename of unformatted binary file
    #Returns a numpy array of data
    #@dx_pts is an integer describing the number of longitude points
    #@dy_pts is an integer describing the number of latitude points
    #@header is a binary variable -- True if contains a 1-elmnt header/footer at 
    #beginning and end of file; else False


    f = open ( infile , 'rb' )
    arr = np.fromfile ( f , dtype = np.float32 , count = -1 )
    f.close()

    if (header == True):
        arr = arr[1:-1]


    #print arr.shape
    #Rebroadcast the data into 2-d array with proper format
    arr.shape = (dy_pts, dx_pts)
    return arr


#Returns an array containing (y,x) grid points in the mask AND
#an array containing (y,x) grid points outside the mask
#mask_var is a np array of binary mask points (1 in the mask; else 0)
#nx is total number of x grid points, ny is total num of y grid points
def get_domain_pts(mask_var, nx, ny):

    mask_pts = []
    out_pts = []
    for i in range(nx):
        for j in range(ny):
            if (mask_var[j,i] >= 0.9):
                mask_pts.append((j,i))
            else:
                out_pts.append((j,i)) 

    mask_pts = np.array(mask_pts)
    out_pts = np.array(out_pts) 

    return mask_pts, out_pts





#Thresholds the forecast at given probability levels. 
#Useful for, e.g., thresholding the RF probs to match the SPC outlooks
#Returns the thresholded forecast values
#The list of thresholds (i.e., thresh_list) should be given in order from largest to smallest prob thresholds
def threshold_forecast(in_probs, thresh_list):
    new_probs = in_probs.copy()

    for t in range(len(thresh_list)):

        if (t == 0):
            new_probs[new_probs >= thresh_list[t]] = thresh_list[t]

            #set everything less than 5% to 0
            new_probs[new_probs <= thresh_list[-1]] = 0.

        else:
            new_probs[(new_probs >= thresh_list[t]) & (new_probs < thresh_list[t-1])] = thresh_list[t]
    
    return new_probs 

#NOTE: For now assumes LCC map projection
#prints image of probs and contoured obs of given forecast (e.g., spc or rf-derived) 
#TODO: Implement plotting individual SPC storm reports as dots with correct color coding 
#the coords variables are np arrays in form [lon, lat]
def get_image(nx, ny, mask_pts, probs_2d, obs_2d, outfile, sw_lon, sw_lat, ne_lon, ne_lat, tlat1, tlat2, stlon, deltax, nonsig_torn_coords, nonsig_wind_coords, nonsig_hail_coords, sig_torn_coords, sig_wind_coords, sig_hail_coords):

    #now can make the plot
    fig = plt.figure(figsize=(12,12), facecolor='whitesmoke')
    Map = Basemap(projection='lcc', llcrnrlon=sw_lon,\
                llcrnrlat=sw_lat,urcrnrlon=ne_lon,urcrnrlat=ne_lat, \
                lat_1 = tlat1, lat_2 = tlat2, lon_0 = stlon, \
                resolution='l')

    y = np.arange(ny)
    x = np.arange(nx)

    xx, yy = np.meshgrid(x,y)

    rng = np.array([0.049999, 0.149999, 0.299999, 0.449999, 0.599999, 0.749999, 1.0])
    #rng = np.array([0.01, 0.02, 0.05, 0.10, 0.15, 0.20, 0.30, 0.40, 0.45, 0.50, 0.60, 0.70, 0.80, 0.90, 1.0]) 
    #rng = np.array([0.02, 0.05, 0.10, 0.15, 0.30, 0.45, 0.60, 0.75, 1.0]) 
    #rng = np.array([0.05, 0.15, 0.30, 0.45, 0.60, 0.75, 1.0])
    rng_obs = np.arange(0., 2., 1.)
    my_cmap = matplotlib.cm.rainbow
    Map.drawcoastlines(linewidth=1.5,color='sienna')
    Map.drawstates(color='sienna', linewidth=1.5)
    Map.drawcountries(linewidth=1.5,color='sienna')
    filled_plt = Map.contourf(xx*deltax, yy*deltax, probs_2d, rng, cmap = my_cmap )
    contour_plt = Map.contour(xx*deltax, yy*deltax, obs_2d, rng_obs, colors='black', linewidths=2)
    cblabel = "Probability of Severe (Any Hazard)"
    cb = fig.colorbar(filled_plt, orientation = 'horizontal', label = cblabel, pad=0.02)
    cb.ax.tick_params(labelsize=16)
    cb.set_label(label = cblabel, size=16) 

    
    #Add storm report "dots"


    #Nonsig 
    add_lsrs(Map, nonsig_hail_coords, "o", "green", 5, False)
    #add_lsrs(Map, nonsig_wind_coords, "o", "blue", 5, False)
    #add_lsrs(Map, nonsig_torn_coords, "o", "red", 5, False)

    #Sig 
    add_lsrs(Map, sig_hail_coords, "^", "black", 7, True)
    #add_lsrs(Map, sig_wind_coords, "s", "black", 7, True)
    #add_lsrs(Map, sig_torn_coords, "o", "red", 7, True)


    #OLD: 
    #Tornado
    #add_lsrs(Map, torn_coords, "o", "red", 5) 

    #Wind 
    #add_lsrs(Map, nonsig_wind_coords, "o", "blue", 5) 
    #add_lsrs(Map, sig_wind_coords, "s", "black", 5)

    #Hail 
    #add_lsrs(Map, nonsig_hail_coords, "o", "green", 5) 
    #add_lsrs(Map, sig_hail_coords, "^", "black", 5)  


    #restrict the plot domain 
    #ymin=43; ymax=190
    #xmin=76; xmax=328

    ymin=min(mask_pts[:,0]*deltax)
    ymax=max(mask_pts[:,0]*deltax)
    xmin=min(mask_pts[:,1]*deltax)
    xmax=max(mask_pts[:,1]*deltax)

    plt.xlim([xmin,xmax])
    plt.ylim([ymin,ymax])

    plt.savefig(outfile)

#Adds appropriate storm report "dots" to a basemap object
#coords_set is a np array where first col is lons, second col is lats
#marker_type is "O" for circle (nonsig), "^" for triangle (sig hail), "s" for square (sig wind) 
#color is red for torn, green for hail, blue for wind
#marker_size is how big to make each marker
#sig is a boolean variable ; True if sig, else False
def add_lsrs(Map_obj, coords_set, marker_type, marker_color, marker_size, sig):

    #Make sure coords_set is shaped correctly
    if (len(coords_set) > 0):
        coords_set.shape = (-1, 2)
        for i in range(len(coords_set)):
            curr_lon = coords_set[i,0]
            curr_lat = coords_set[i,1]

            if (sig == False):
                Map_obj.plot(Map_obj(curr_lon, curr_lat)[0], Map_obj(curr_lon, curr_lat)[1], marker=marker_type, color=marker_color, markersize=marker_size, markeredgecolor='black', markeredgewidth=1.0)

            else:
                Map_obj.plot(Map_obj(curr_lon, curr_lat)[0], Map_obj(curr_lon, curr_lat)[1], marker=marker_type, color=marker_color, markersize=marker_size, markeredgecolor='white', markeredgewidth=1.0)

    return


def getAggStats(probabilities, Y_eval):
    #numForecasts is the number of forecasts in each probability bin
    RELx, RELy, numForecasts = Eval.ReliabilityCurve( probabilities, Y_eval)
    POFD, POD = Eval.rocCurve(probabilities, Y_eval )

    JUNK, SR = Eval.perfDiag(probabilities, Y_eval)
    AUC = roc_auc_score(Y_eval, probabilities)

    BSS, BS = Eval.bss_no_train(Y_eval, probabilities)

    BSS_rel, BSS_res, BSS_unc = Eval.BS_decomposition(Y_eval, probabilities)

    return POD, POFD, SR, AUC, RELx, RELy, BSS, BSS_rel, BSS_res, BSS_unc, BS, numForecasts


#=========================

#==========================
# User defined parameters
#==========================

dx = 93
dy = 65

#ddir_rf = "/glade/scratch/eloken/sseo/spc_outlooks/kml_files/rf_probs_full_new_obs_new_baseline"
ddir_rf = "./outfiles"

#Directories to storm report coordinates (lon, lat) 
ddir_sig_coords = "/glade/work/eloken/storm_reports/new/coords/sig"
ddir_nonsig_coords = "/glade/work/eloken/storm_reports/new/coords/nonsig" 

fdate_file = "test_dates.txt"

#obs_dir = "/glade/work/eloken/storm_reports/new/all_hazards/80km"

obs_dir = "../all_y"

maskfile = "new_conus_mask_80km.dat"

#NCEP 211 stats
deltax = 81.2705*10**3 # horizontal grid spacing in m; 81.2705 km
sw_lon = -133.459
sw_lat = 12.190
ne_lon = -49.385
ne_lat = 57.290
tlat1 = 25.0 #lat1
tlat2 = 25.0 #lat2 
stlon = -95.0  #lon_0

#===========================

#Read in the testing dates
fdates = np.genfromtxt(fdate_file, dtype='str') 

#fdates = ["20150526", "20150527"] 

#fdates = fdates[0:3] #for testing/debugging

#TODO: Read in/obtain mask points
domain_mask = read_binary(maskfile, dx, dy, True)

#Obtain x and y points/coordinates that fall within the domain
domain_pts, non_domain_pts = get_domain_pts(domain_mask, dx, dy)

#non-domain x- and y-points
ypts = non_domain_pts[:,0]
xpts = non_domain_pts[:,1]

overall_rf = []
overall_obs = []
for d in range(len(fdates)):
    fdate = fdates[d]
    print (fdate) 
    #Read in the probabilities 

    rf_probs = np.genfromtxt("%s/%s_rf_probs.txt" %(ddir_rf, fdate),dtype=float ) 


    #Threshold the rf_probs to match the spc probs
    #NOTE: Need to put this back in to be fair. 
    #rf_probs = threshold_forecast(rf_probs, thresholds) 

    #Read in the observations
    daily_obs = read_binary("%s/%s00_REPS211_hail.dat" %(obs_dir, fdate), dx, dy, False)
   
    #Read in the lons/lats for each hazard type (torn, wind, hail; sig and nonsig ) 
    coords_nonsig_torn = np.genfromtxt("%s/nonsig_tor_coords_%s.txt" %(ddir_nonsig_coords, fdate), dtype='float' )
    coords_nonsig_wind = np.genfromtxt("%s/nonsig_wind_coords_%s.txt" %(ddir_nonsig_coords, fdate), dtype='float')
    coords_nonsig_hail = np.genfromtxt("%s/nonsig_hail_coords_%s.txt" %(ddir_nonsig_coords, fdate), dtype='float')


    #TODO: Add sig tor coordinates; also make sig images have white outline
    coords_sig_wind = np.genfromtxt("%s/sig_wind_coords_%s.txt" %(ddir_sig_coords, fdate), dtype='float')
    coords_sig_hail = np.genfromtxt("%s/sig_hail_coords_%s.txt" %(ddir_sig_coords, fdate), dtype='float')
    coords_sig_torn = np.genfromtxt("%s/sig_tor_coords_%s.txt" %(ddir_sig_coords, fdate), dtype='float') 


    #Filter out points outside of the analysis domain -- step 1: assign all non-domain points "NaN"  --
    #These points will be filtered more below once they are put in 1-d form 
    rf_probs[ypts, xpts] = np.nan
    daily_obs[ypts, xpts] = np.nan

    #Get 2d images in basemap

    #if (fdate == "20170527" or fdate == "20150509" or fdate == "20160315" or fdate == "20150622" or fdate == "20150425"):
    #if (fdate == "20150509"):
    #get_image(dx, dy, domain_pts, rf_probs, daily_obs, "%s_rf_probs_severe.png" %fdate, sw_lon, sw_lat, ne_lon, ne_lat, tlat1, tlat2, stlon, deltax, coords_nonsig_torn, coords_nonsig_wind, coords_nonsig_hail, coords_sig_torn, coords_sig_wind, coords_sig_hail)

    #Get raw baseline images and add to overall/aggregated bucket

    #Now work on computing stats

    #Add probs to an overall/aggregated bucket

    [overall_rf.append(x) for x in rf_probs.flatten() if str(x) != 'nan']
    [overall_obs.append(x) for x in daily_obs.flatten() if str(x) != 'nan']

#At the end, should be able to compute overall stats
overall_rf = np.array(overall_rf)
overall_obs = np.array(overall_obs) 

#print ("overall uh original" ) 
#print (overall_uh) 

climo_freq = sum(overall_obs)/len(overall_obs) 

POD_rf, POFD_rf, SR_rf, AUC_rf, RelX_rf, RelY_rf, BSS_rf, BS_rel_rf, BS_res_rf, BS_unc_rf, BS_rf, n_fcsts_rf = getAggStats(overall_rf, overall_obs)



print("RF stats:")
print ("AUC: ", AUC_rf)
print ("BS: ", BS_rf)
print ("BSS: ", BSS_rf)
print ("BS_rel: ", BS_rel_rf)
print ("BS_res: ", BS_res_rf)
print ("BS_unc: ", BS_unc_rf)

#produce the graphs 
#perf diagram
#reliability
#auc

#RF in red, spc in blue

#########################
# Make Attributes plot
#########################

RelXs = [RelX_rf] 
RelYs = [RelY_rf] 
n_fcsts = [n_fcsts_rf]
SRs = [SR_rf]
PODs = [POD_rf]
POFDs = [POFD_rf]
AUCs = [AUC_rf]
BSSs = [BSS_rf]
BS_rels = [BS_rel_rf]
BS_reses = [BS_res_rf]


#possible_colors = ["lightpink", "mediumpurple", "darkviolet", "aqua", "navy", "lime", "darkgreen", "gold", "darkorange", "firebrick", "black" ]
possible_colors = ["darkviolet", "mediumpurple", "aqua", "navy", "lime", "darkgreen", "gold", "darkorange", "firebrick", "black" ]

col_array = ["red"] 
label_array = ["RF"] 
label_array_orig = ["RF"] 

#n_fcsts = [n_fcsts_rf, n_fcsts_spc]

fig = plt.figure(figsize = (12, 12), facecolor = 'whitesmoke')
ax = fig.add_subplot(1,1,1)

graph.makeAttributes(RelXs, RelYs, col_array, label_array_orig, climo_freq, n_fcsts, ax) 

plt.savefig("attributes_severe.png")

#############################
#Make performance diagram
#############################

#SRs = [SR_rf, SR_spc]
#PODs = [POD_rf, POD_spc] 

fig = plt.figure(figsize = (12,12), facecolor='whitesmoke') 
ax = fig.add_subplot(1,1,1) 

graph.makePerfDiag(SRs, PODs, col_array, label_array_orig, ax) 

plt.savefig("performance_severe.png") 


###########################
# Make ROC diagram
###########################

#POFDs = [POFD_rf, POFD_spc]
#PODs = [POD_rf, POD_spc]
#AUCs = [AUC_rf, AUC_spc]

fig = plt.figure(figsize = (12,12), facecolor='whitesmoke')
ax = fig.add_subplot(1,1,1)

graph.makeROC(POFDs, PODs, AUCs, col_array, label_array_orig, ax) 

plt.savefig("roc_severe.png") 



#########################
# Make bar plots
#########################

#def makeBarPlots(AUCs_sref, AUCs_href, col_arr, lab_arr, ens_arr, ymin, ymax, ylab, isLog, isSciNot, ax)
#multi_dim_barplot(aucs, bsses, bs_rels, bs_reses, col_arr, lab_arr, ax)

#BSSs = [BSS_rf, BSS_spc]
#BS_rels = [BS_rel_rf, BS_rel_spc]
#BS_reses = [BS_res_rf, BS_res_spc] 

fig = plt.figure(figsize = (12,12), facecolor='whitesmoke')
ax = fig.add_subplot(1,1,1)

#graph.multi_dim_barplot(AUCs, BSSs, BS_rels, BS_reses, col_array, label_array, ax) 


#AUC
#graph.single_dim_barplot(AUCs, col_array, label_array, 0.50, 1.0, "AUC", False,  ax)

#plt.savefig("barPlots_auc.png" ) 

#BSS

#fig = plt.figure(figsize = (12,12), facecolor='whitesmoke')
#ax = fig.add_subplot(1,1,1)

#graph.single_dim_barplot(BSSs, col_array, label_array, -0.10, 0.25, "BSS", False, ax)
#plt.subplots_adjust(left=0.15) 
#plt.savefig("barPlots_bss.png" )


#BSS Rel


#fig = plt.figure(figsize = (12,12), facecolor='whitesmoke')
#ax = fig.add_subplot(1,1,1)

#graph.single_dim_barplot(BS_rels, col_array, label_array, 0.0, 0.001, "BS Rel.", True, ax)

#plt.subplots_adjust(left=0.15) 
#plt.savefig("barPlots_bs_rel.png" )



#BS Res.

#fig = plt.figure(figsize = (12,12), facecolor='whitesmoke')
#ax = fig.add_subplot(1,1,1)

#graph.single_dim_barplot(BS_reses, col_array, label_array, 0.0, 0.0045, "BS Res.", True, ax)


#plt.subplots_adjust(left=0.15) 

#plt.savefig("barPlots_bs_res.png" )





